
/*
 * Calculate positions of Jupiter's moons.
 *
 * Equations come from Jean Meeus, Astronomical Formulae for Calculators.
 *
 *	Copyright 1996, 2002 by Akkana Peck.
 */

#include "shallow.h"

/*
 * Current system II longitude of the great red spot, in degrees.
 * This changes from year to year; update it here.
 * It's compiled in to the program to avoid having to keep a data
 * file on the palm just for that.
 */
#define GRS_LONG 89

#define NUM_MOONS 4

#ifdef PALMOS
#define INFINITY 1e99
#else
#define INFINITY HUGE_VAL
#endif

/* Globals that other files need */
double julianDate;
int DrawOrbits = 0;

/* Angles of each of the Galilean satellites, in radians,
 * expressed relative to each satellite's inferior conjunction:
 */
static double moonAngles[NUM_MOONS];
/* And their distances from the planet: */
static double moonDist[NUM_MOONS];
/* And their abbreviations: */
static char* moonLabels[NUM_MOONS] = {"I", "E", "G", "C" };
    
static double delta;	/* Earth-Jupiter distance */
static double De;	/* planetocentric ang. earth dist. from jup. equator */
static double G, H;

/* latitudes of systems I and II: */
static double lambda1;
static double lambda2;

typedef struct { double x; double y; } XYCoord;
    
/* variables we may want later for moon calcs: */
double psi;

/*
 * Convert an angle (in radians) so that it's between 0 and 2*PI:
 */
double angle(double a)
{
    return fmod(a, (double)(2.*PI));
}

void setDate(double jd)
{
    /* Calculate the position of Jupiter's central meridian: */
	
    /* First, get the number of days since 1899 Dec 31 12h ET. */
    double d = jd - 2415020;     /* days since 1899 Dec 31 12h ET */
	
    /* Argument for the long-period term in the motion of Jupiter: */
    double V = angle((134.63 + .00111587 * d) * PI / 180);
	
    /* Mean anomalies of Earth and Jupiter: */
    double M = angle((358.476 + .9856003 * d) * PI / 180);
    double N = angle((225.328 + .0830853 * d + .33 * sin(V))
                     * PI / 180);
	
    /* Diff between the mean heliocentric longitudes of Earth & Jupiter: */
    double J = angle((221.647 + .9025179 * d - .33 * sin(V))
                     * PI / 180);
	
    /* Equations of the center of Earth and Jupiter: */
    double A = angle((1.916 * sin(M) + .020 * sin(2*M))
                     * PI / 180);
    double B = angle((5.552 * sin(N) + .167 * sin(2*N))
                     * PI / 180);
	
    double K = angle(J + A - B);
	
    /* Distances are specified in AU: */
    /* Radius vector of the earth: */
    double R = 1.00014 - .01672 * cos(M) - .00014 * cos(2*M);
    /* Radius vector of Jupiter: */
    double r = 5.20867 - .25192 * cos(N) - .00610 * cos(2*N);

    double lambda;
	
    /* Earth-Jupiter distance: */
    delta = sqrt(r*r + R*R - 2*r*R*cos(K));
	
    /* Phase angle of Jupiter (always btw. -12 and 12 degrees): */
    psi = asin(R / delta * sin(K));
	
    /* Longitude of system 1: */
    lambda1 = angle((268.28 * 877.8169088 * (d - delta / 173))
                    * PI / 180 + psi - B);
    /* Longitude of system 2: */
    lambda2 = angle((290.28 + 870.1869088 * (d - delta / 173))
                    * PI / 180 + psi - B);

    /* calculate the angles of each of the satellites: */
    moonAngles[0] = angle((84.5506 + 203.4058630 * (d - delta / 173))
                          * PI / 180
                          + psi - B);
    moonAngles[1] = angle((41.5015 + 101.2916323 * (d - delta / 173))
                          * PI / 180
                          + psi - B);
    moonAngles[2] = angle((109.9770 + 50.2345169 * (d - delta / 173))
                          * PI / 180
                          + psi - B);
    moonAngles[3] = angle((176.3586 + 21.4879802 * (d - delta / 173))
                          * PI / 180
                          + psi - B);
	
    /* and the planetocentric angular distance of the earth
     * from the equator of Jupiter:
     */
    lambda = angle((238.05 + .083091 * d + .33 * sin(V))
                   * PI / 180 + B);
    De = ((3.07 * sin(lambda + 44.5 * PI / 180)
           - 2.15 * sin(psi) * cos(lambda - 24.*PI/180)
           - 1.31 * (r - delta) / delta
           * sin(lambda - 99.4 * PI / 180))
          * PI / 180);
	
    G = angle((187.3 + 50.310674 * (d - delta / 173)) * PI / 180);
    H = angle((311.1 + 21.569229 * (d - delta / 173)) * PI / 180);

    /* Calculate the distances before any corrections are applied: */
    moonDist[0] = 5.9061 -
        .0244 * cos(2 * (moonAngles[0] - moonAngles[1]));
    moonDist[1] = 9.3972 -
        .0889 * cos(2 * (moonAngles[1] - moonAngles[2]));
    moonDist[2] = 14.9894 - .0227 * cos(G);
    moonDist[3] = 26.3649 - .1944 * cos(H);
	
    /* apply some first-order correction terms to the angles: */
    moonAngles[0] = angle(moonAngles[0] +
                          sin(2 * (moonAngles[0] - moonAngles[1]))
                          * .472 * PI / 180);
    moonAngles[1] = angle(moonAngles[1] +
                          sin(2 * (moonAngles[1] - moonAngles[2]))
                          * 1.073 * PI / 180);
    moonAngles[2] = angle(moonAngles[2] +
                          sin(G) * .174 * PI / 180);
    moonAngles[3] = angle(moonAngles[3] +
                          sin(H) * .845 * PI / 180);
}
    
/*
 * Routines to get distances, angles, or X and Y coordinates
 * for each of the moons.
 * These assume that moon positions have already been calculated!
 */
double getMoonDist(int whichmoon)	{ return moonDist[whichmoon]; }
double getMoonAngle(int whichmoon)	{ return moonAngles[whichmoon]; }
char* getMoonLabel(int whichmoon)	{ return moonLabels[whichmoon]; }

void getMoonXY(XYCoord* coord, int whichmoon)
{
    double r = getMoonDist(whichmoon);

    /* See whether the moon is hidden behind the planet: */
    coord->x = -r * sin(moonAngles[whichmoon]);
    coord->y = -r * cos(moonAngles[whichmoon]) * sin(De);

    if (coord->x < 1. && coord->x > -1.
        && moonAngles[whichmoon] > PI * .5
        && moonAngles[whichmoon] < PI * 1.5)
    {
        coord->x = coord->y = INFINITY;
        /* NOTE!  Can't use NaN, because comparisons with !=
         * don't work against NaN.  (In Java, anyway.)
         */
    }
}
    
/*
 * Moon shadows
 */
void getMoonShadowXY(XYCoord* coord, int whichmoon)
{
    double r = getMoonDist(whichmoon);
    double moonSunAngle = moonAngles[whichmoon] - psi;
	
    coord->x = -r * sin(moonSunAngle);
    /* This Y coord isn't right ... need to derive the right eqn: */
    coord->y = -r * cos(moonSunAngle) * sin(De);

    /* See whether the shadow is on the far side of the planet: */
    if (coord->x > -1. && coord->x < 1.
        && moonSunAngle > PI * .5 && moonSunAngle < PI * 1.5)
    {
        coord->x = coord->y = INFINITY;
    }
    else if (coord->x < -1. || coord->x > 1.)
    {
        /* shadow isn't hitting the planet right now.
         * Some day, ought to check for moons eclipsing other moons.
         */
        coord->x = coord->y = INFINITY;
    }
}
    
/*
 * The Red Spot
 */
void getRedSpotXY(XYCoord* coord)
{
    double spotlong = angle(lambda2 - GRS_LONG*PI/180);
	
    /* See if the spot is visible: */
    if ((spotlong > (PI * .5)) && (spotlong < (PI * 1.5))) {
        coord->x = coord->y = INFINITY;
    } else {
        coord->x = -sin(spotlong);
        coord->y = -.40;	/* approximate guess */
    }
}

#define BAND_DX(i) (radius*.98 - ((i)-bandpos)/2.3)

void EraseOutside(int x, int y, int width, int height,
                  int radius, int bandpos, int XC, int YC,
                  int flipNS, int flipEW,
                  RGBColorType* color)
{
    int endx, i;
    int istop = (y < YC);
    int isleft = (x < XC);

    ChangeColor(color);
    width *= 2;
    endx = x + width * (isleft ? -1 : 1);
    for (i = -height; i <= height; ++i)
    {
        int yy = i + y;
        int xx = XC + ((2 + (BAND_DX((yy-YC) * (flipNS^istop ? -1 : 1))))
                       * (isleft ? -1 : 1));
        if ((isleft && (endx < xx)) || (!isleft && (endx > xx)))
            Line(endx, yy, xx, yy);
    }
}

#define SPOTASP .65

int labelCollides(int x, int y, int i)
{
#define MINY 10
#define charSlopX 7
#define charSlopY 10
    int j;
    static int labelsX[NUM_MOONS-1];
    static int labelsY[NUM_MOONS-1];

    labelsX[i] = x;
    labelsY[i] = y;

    for (j=0; j<i && y > MINY; ++j)
    {
        if ((x > labelsX[j]-charSlopX && x < labelsX[j]+charSlopX)
            &&
            (y > labelsY[j]-charSlopY && y < labelsY[j]+charSlopY))
            return 1;
    }
    return 0;
}

/* Draw Jupiter.
 * If color is off, we'll ignore spot/band/disk colors.
 */
void DrawJupiter(int Xmax, int Ymax, double Scale,
                 int flipNS, int flipEW, int color,
                 RGBColorType* fgC, RGBColorType* bgC,
                 RGBColorType* spotC, RGBColorType* bandC, RGBColorType* diskC)
{
    int XC = Xmax / 2;
    int YC = Ymax / 2;
    int radius = Scale*Ymax;

    int bandpos = radius/4;
    int bandheight = radius/3.5;
    /*int EbandSpace = radius/6.9;*/
    int spotsize = radius/5 + 2;
    int moonsize = radius/12;

    int dband = color ? 1 : 2;
    int i;
    int temp;

    XYCoord coord;

    setDate(julianDate);

    ChangeColor(fgC);

    /* Label the cardinal directions: */
    DrawString(XC,      10,     flipNS ? "-N-" : "-S-", 3);
    DrawString(XC,      Ymax-4, flipNS ? "-S-" : "-N-", 3);
    DrawString(Xmax-15, YC,     flipEW ? "-W"  : "-E",  2);
    DrawString(3,       YC,     flipEW ?  "E-" :  "W-", 2);

    if (color) {
        ChangeColor(diskC);
        Oval(XC, YC, radius, 1., 1);
    }
    else {
        Oval(XC, YC, radius, 1., 0);
    }

    /* Draw bands */
    if (color)
        ChangeColor(bandC);
    for (i = bandpos; i <= bandpos + bandheight; i += dband)
    {
        int dx = BAND_DX(i);
            
        Line(XC-dx, YC-i, XC+dx, YC-i);
        Line(XC-dx, YC+i, XC+dx, YC+i);
    }

    /* Draw spot */
    getRedSpotXY(&coord);
    if (coord.x != INFINITY && coord.y != INFINITY)
    {
        int x = XC + coord.x * radius * (flipEW ? -1 : 1);
        int y = YC + coord.y * radius * (flipNS ? -1 : 1);
        int overlap = (double)radius*.5;

        /* outline first */
        ChangeColor(fgC);
        Oval(x, y , spotsize, SPOTASP, 1);
        /* shade the spot */
        if (color) {
            ChangeColor(spotC);
            Oval(x, y, spotsize-1, SPOTASP, 1);
        }

        /* Chop off the outside of the spot if necessary. */
        if (x + spotsize > XC + overlap)        /* hangs off the right */
            EraseOutside(x, y, spotsize, spotsize*SPOTASP,
                         radius, bandpos, XC, YC,
                         flipNS, flipEW, bgC);
        else if (x - spotsize < XC - overlap)   /* hangs off the left */
            EraseOutside(x, y, spotsize, spotsize*SPOTASP,
                         radius, bandpos, XC, YC,
                         flipNS, flipEW, bgC);
    }

    /* Draw the moons: */
    if (moonsize < 2)
        moonsize = 2;
    for (i=0; i<NUM_MOONS; ++i)
    {
        /* See if this moon is casting a shadow: */
        getMoonShadowXY(&coord, i);
        if (coord.x != INFINITY && coord.y != INFINITY)
        {
            int x = XC + coord.x * radius * (flipEW ? -1 : 1);
            int y = YC + coord.y * radius * (flipNS ? -1 : 1);
            if (color)
                ChangeColor(bgC);
            else
                ChangeColor(fgC);
            Oval(x, y, moonsize+1, 1, 1);

            /* If the shadow is near the edge, it might spill over
             * in non-color mode, so chop it off:
             */
            if (!color)
            {
                if (x + moonsize + radius >= XC)
                    EraseOutside(x, y, moonsize, moonsize, radius, bandpos,
                                 XC, YC, flipNS, flipEW, bgC);
                else if (x - moonsize - radius <= XC)
                    EraseOutside(x, y, moonsize, moonsize, radius, bandpos,
                                 XC, YC, flipNS, flipEW, bgC);
            }

            /* Figure out where to draw the label.
             * XXX This may not end up being visible.
             */
            temp = YC-radius*2.5 + coord.y * radius * (flipNS ? -1 : 1);
            if (temp > 10) {
                y = temp;
                if (color)
                    ChangeColor(diskC);
                else
                    ChangeColor(fgC);
            }
            else {
                x += moonsize;
                ChangeColor(fgC);
            }
            DrawString(x, y, moonLabels[i], 1);
        }

        /* Draw the moon. */
        getMoonXY(&coord, i);
        if (coord.x != INFINITY && coord.y != INFINITY)
        {
            int x = XC + coord.x * radius * (flipEW ? -1 : 1);
            int y = YC + coord.y * radius * (flipNS ? -1 : 1);

            /* outline first if it's against the disk. */
            if (coord.x >= -1 && coord.x <= 1)
            {
                ChangeColor(bgC);
                if (!color) {
                    Oval(x, y, moonsize, 1, 1);
                    ChangeColor(fgC);
                }
                Oval(x, y, moonsize+1, 1, 0);
            }

            /* fill the moon itself, inside the outline */
            ChangeColor(fgC);
            Oval(x, y, moonsize, 1, color ? 1 : 0);

            /* Label the moon */
            temp = YC-radius*2.5 + coord.y * radius * (flipNS ? -1 : 1);
            if (temp > 10) {
                y = temp;
            }
            else {
                y -= moonsize*2;
                x += moonsize*2;
                if (labelCollides(x, y, i))
                {
                    x -= moonsize*2+charSlopX;
                    if (labelCollides(x, y, i))
                        y += moonsize*2 + charSlopY;
                }
            }
            DrawString(x, y, moonLabels[i], 1);
        }
    }
}

