/*
 * palmgrfx.c
 * Graphics and time routines for planetary programs on PalmOS.
 *
 * Copyright 2002 by Akkana Peck.
 * You are free to use or modify this code under the Gnu Public License.
 */

#include "shallow.h"
#include "palmgrfx.h"

RGBColorType white = {255, 255, 255};
RGBColorType black = {  0,   0,   0};

static int greyMode = true;

void ChangeColor(RGBColorType* c)
{
    RGBColorType old;

    // This is deprecated and left in for pre-3.5 roms
    // but actually should work anywhere.
    // XXX Except it doesn't work in satmoons, tho it does in marsmap.
    if (greyMode)
        WinSetColors(c, &old, &white, &old);
    else
        WinSetColors(&black, &old, &white, &old);
    return;

#ifdef PALM_APIS_CHANGE_TOO_FREQUENTLY_TO_KEEP_UP
    RGBColorType standin = dark;

#if defined(SDK_31)  /* things are so simple with SDK 3.1! */
    if (greyMode)
        WinSetForeColor(c, &old);
    else
        WinSetForeColor(&standin, &old);
#endif

#if defined(SDK_40)
    if (romVersion > sysVersion35)
    {
        if (greyMode)
            WinSetForeColorRGB(c, &old);
        else
            WinSetForeColorRGB(&standin, &old);
        return;
    }
#endif

#if defined(SDK_35) || defined(SDK_40)
    if (romVersion >= sysVersion35)
    {
        IndexedColorType indexC, oldC;
        if (greyMode)
            indexC = WinRGBToIndex(c);
        else
            indexC = WinRGBToIndex(&standin);
        oldC = WinSetForeColor(indexC);
        return;
    }
#endif
#endif /* PALM_APIS_CHANGE_TOO_FREQUENTLY_TO_KEEP_UP */
}

void ClearScreen()
{
    WinEraseWindow();
}

void ClearArea(int x1, int y1, int x2, int y2)
{
    RectangleType rect;
    rect.topLeft.x = x1;
    rect.topLeft.y = y1;
    rect.extent.x = x2;
    rect.extent.y = y2;
    ChangeColor(&white);
    WinFillRectangle(&rect, 1);
}

void DrawString(int x, int y, char* s, int len)
{
    WinDrawChars(s, len, x, y-9);
}

void Line(int x1, int y1, int x2, int y2)
{
    WinDrawLine(x1, y1, x2, y2);
}

void Oval(int xc, int yc, int radius, double aspect, int fill)
{
    int i;
    int r2;
    int lasty;

    // if it's very small, just draw a little diamond
    if (radius <= 2 && !fill)
    {
        radius = 2;
        Line(xc-radius, yc, xc, yc-radius);
        Line(xc, yc-radius, xc+radius, yc);
        Line(xc+radius, yc, xc, yc+radius);
        Line(xc, yc+radius, xc-radius, yc);
        return;
    }
    // Otherwise draw a real circle
    r2 = radius*radius;
    lasty = -1;
    for (i = 0; i <= radius; ++i)    // loop over X from center to right
    {
        int yr = sqrt(r2 - (double)i*i) * aspect;
        if (lasty < 0) lasty = yr;
        if (fill)
        {
            Rectangle(xc-i, yc-yr, i*2, yr*2, 1);
            //Rectangle(xc-i, yc-yr, i*2, lasty-yr+1, 1);
            // Next one isn't right:
            //Rectangle(xc-i, yc, i*2, yr, 1);
        }
        else
        {
            Line(xc-i, yc-lasty, xc-i, yc-yr);
            Line(xc-i, yc+lasty, xc-i, yc+yr);
            Line(xc+i, yc-lasty, xc+i, yc-yr);
            Line(xc+i, yc+lasty, xc+i, yc+yr);
        }
        lasty = yr;
    } // end loop over X
}

void Rectangle(int x, int y, int width, int height, int fill)
{
    RectangleType rect;
    rect.topLeft.x = x;
    rect.topLeft.y = y;
    rect.extent.x = width;
    rect.extent.y = height;
    if (fill)
        WinDrawRectangle(&rect, 1);
    else
        WinDrawRectangleFrame(roundFrame, &rect);
}

void Arc(int xc, int yc, int r, double a1, double a2, double aspect, int fill)
{
    /* not implemented */
}

double PalmTimeToJulian(unsigned long int sec)
{
    /* sec is seconds since beginning of 1904-01-01 UTC */
    return(sec/86400.0 + 2416480.498) - TZadjustment;
}

unsigned long JulianToPalmTime(double julian)
{
    return (julian + TZadjustment - 2416480.498) * 86400;
}

char* JulianToString(double julian)
{
    static char Buffer[200];
    unsigned long nowseconds = JulianToPalmTime(julianDate);
    DateTimeType datetime;
    TimSecondsToDateTime(nowseconds, &datetime);
    StrPrintF(Buffer, "%2d:%02d %2d/%2d/%-4d TZ %d  ",
              datetime.hour, datetime.minute,
              datetime.month, datetime.day, datetime.year, (int)TimeZone);
    return Buffer;
}

/* ErrFatalDisplayIf always crashes visors and Clies, so don't use it.
 * Instead, call FrmCustomAlert with the message.
 * We can't do that here because it depends on a #define
 * that's generated by pilrc.
 */
extern void FatalError(char* s);

void AppLoop()
{
    short err;
    EventType event;

    do {
        EvtGetEvent(&event, 100);
        if (!SysHandleEvent(&event)) 
            if (!MenuHandleEvent((void *)0, &event, &err)) 
                if (event.eType != appStopEvent) {
                    if (FrmGetActiveForm())
                        FrmDispatchEvent(&event);
                    else
                        HandleEvent(&event);
                }
    } while (event.eType != appStopEvent);
}

Err StartApp(const Word mainForm)
{
    Err error;
    DWord depth = 2;

#if defined(SDK_40)
    if (SysGetTrapAddress (sysTrapWinSetForeColorRGB)
        == SysGetTrapAddress (sysTrapSysUnimplemented))
        greyMode = false;
    else
#endif /* SDK_40 */
    {
        error = ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &depth, NULL);
        if (!error) {
            //SetGreyscalePalette(0, 2, 3, 7);
            greyMode = true;
        } else
            greyMode = false;
    }

    FrmGotoForm(mainForm);

    error = SysLibFind(MathLibName, &MathLibRef);
    if (error)
        error = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
    if (!error) MathLibOpen(MathLibRef, MathLibVersion);
    if (error) {
        FatalError("Can't find MathLib");
    }

    TimeZone = GetTimeZoneOffset();
    TZadjustment = TimeZone / 24.0;
    julianDate = PalmTimeToJulian(TimGetSeconds());

    return(0);
}

Err StopApp(void)
{
    UInt usecount;
    Err error;

    error = MathLibClose(MathLibRef, &usecount);
    if (error)
        FatalError("Can't close MathLib");
    if (usecount == 0)
        SysLibRemove(MathLibRef);
    FrmCloseAllForms();
    ScrDisplayMode(scrDisplayModeSetToDefaults, NULL, NULL, NULL, NULL);
    return(0);
}


