/*
 * satmoons.c
 * Calculations of positions of Saturn and its moons.
 * Some of the code was adapted from SATSAT2.BAS
 * by Dan Bruton, astro@tamu.edu.
 * See "Practical Astronomy with your Calculator" by Peter Duffett-Smith,
 * and the Astronomical Almanac for explanations of some of these
 * calculations.
 *
 * Copyright 2002 by Akkana Peck.
 * You may use or modify this code under the Gnu Public License.
 */

#include "shallow.h"

#define SIGN(a)  (((a)>=0) ? 1 : -1)

/* Constants */
long RS = 60330;   /* Radius of planet in kilometers */
long RS1 = 92000;  /* Inner radius of inner ring in kilometers */
/* Fudge RS2 and RS3 so Cassini's division will show more clearly */
long RS2 = 111000; /* Outer radius of inner ring (really 117500) */
long RS3 = 130000; /* Inner radius of outer ring (really 122500) */
long RS4 = 136500; /* Outer radius of outer ring in kilometers */

#define NSAT 10

double julianDate;

int DrawOrbits = 0;

int NumMoons = 8;  /* Number of moons displayed */

/* What we need to know about a planet for this program.
 * These aren't full orbital elements.
 */
typedef struct planet_s {
    double orbitalRate;    /* Orbital rate, radians/day */
    double ecc;            /* eccentricity */
    double sma;            /* semi-major axis */
} planet_t;

static planet_t earth =
{
    TWOPI / (1.00004 * 365.2422),
    .016718,
    1.
};

static planet_t saturn =
{
    TWOPI / (29.45771 * 365.2422),
    .0556155,
    9.554747
};

double JDE = 2444238.5;   /* Epoch Jan 0.0 1980 = December 31,1979 0:0:0 UT */
double LpE = 165.322242 * D2R;  /* Longitude of Saturn at Epoch */
double LpP = 92.6653974 * D2R;  /* Longitude of Saturn`s Perihelion */
double Ohm = 113.4888341 * D2R; /* Longitude of Saturn`s Ascending Node */
double IncS = 2.4893741 * D2R;  /* Inclination of Saturn`s Orbit */
double LEE = 98.83354 * D2R;    /* Longitude of Earth at Epoch */
double LEP = 102.596403 * D2R;  /* Longitude of Earth's Perihelion */
static double Obl = 23.43928 * D2R; /* obliquity of the ecliptic */

/*
 * Kludge alert!
 * This table needs to be reset every year or so.
 * Reset the U0 values, the Angle from inferior geocentric conjuction
 * measured westward along the orbit at epoch.
 * Pick a time at which you want accuracy to be highest
 * (probably opposition or a month or two later),
 * calculate the Julian time, and estimate the moon positions
 * for that time.
 */
//double JDEmoons = 2452275.5;    /* Jan 1.0 2002 */
double JDEmoons = 2452640.5;    /* Jan 1.0 2003 */
satellite_t satMoons[] =
{
/*    Name           SMA   E     Period  mag*10   U0      U */
/*    ----           ---   -     ------  -----  ------    - */
    { "Mimas",     185600, 0,   .9425049, 130, 185 * D2R, 0 },
    { "Enceladus", 238100, 0,  1.3703731, 118, 230 * D2R, 0 },
    { "Tethys",    294700, 0,  1.8880926, 103, 160 * D2R, 0 },
    { "Dione",     377500, 0,  2.7375218, 102, 132 * D2R, 0 },
    { "Rhea",      527200, 0,  4.519163,   98,  97 * D2R, 0 },
    { "Titan",    1221600, 0, 15.9669028 , 84, 137 * D2R, 0 },
    { "Hyperion", 1483000, 0, 21.3174647, 143, 335 * D2R, 0 },
    { "Iapetus",  3560100, 0, 79.9208313, 112, 290 * D2R, 0 },
};

/* Inclination of Rings and other items */
double Inclination, IapInci, IapGam;

/********************************************************************
 *                                                                  *
 *   Orbit Calculations                                             *
 *                                                                  *
 ********************************************************************
 */
void SaturnOrbit(double JD)
{
    int i;
    double lE, lP, rE, rP, dt, ii, f, f1;
    double psi, lpq, rpq, lamb, beta, RA, dec;
    double tva, pva, tvc, pvc, dot1, tvb, pvb, dot2;
    double trip;

    /******************* FIND MOON ORBITAL ANGLES ************************/
    double NN = JD - JDE;   /* NN = Number of days since epoch */
    double ME = ((earth.orbitalRate * NN) + LEE - LEP); /* Mean Anom, Earth */
    double MP = ((saturn.orbitalRate * NN) + LpE - LpP);/* Mean Anom, Saturn */
    double VE = ME;
    double VP = MP;   /* True Anomolies - Solve Kepler's Equation */
    for (i=0; i < 3; ++i)
    {
        VE = VE - (VE - (earth.ecc * sin(VE)) - ME) / (1 - (earth.ecc * cos(VE)));
        VP = VP - (VP - (saturn.ecc * sin(VP)) - MP) / (1 - (saturn.ecc * cos(VP)));
    }

    VE = 2 * atan(sqrt((1. + earth.ecc) / (1. - earth.ecc)) * tan(VE / 2.));
    if (VE < 0) VE = TWOPI + VE;
    VP = 2 * atan(sqrt((1. + saturn.ecc) / (1. - saturn.ecc)) * tan(VP / 2.));
    if (VP < 0) VP = TWOPI + VP;

    /* Heliocentric Longitudes of Earth and Saturn */
    lE = VE + LEP;
    if (lE > TWOPI) lE = lE - TWOPI;
    lP = VP + LpP;
    if (lP > TWOPI) lP = lP - TWOPI;

    /* Distances of Earth and Saturn from the Sun in AU's */
    rE = earth.sma * (1. - earth.ecc * earth.ecc) / (1. + earth.ecc * cos(VE));
    rP = saturn.sma * (1. - saturn.ecc * saturn.ecc)
         / (1. + saturn.ecc * cos(VP));
    /* DT = Distance from Saturn to Earth in AU's - Law of Cosines */
    dt = sqrt((rE * rE) + (rP * rP) - (2. * rE * rP * cos(lE - lP)));

    /* II = Angle between Earth and Sun as seen from Saturn */
    ii = asin(rE * sin(lP - lP) / dt);

    /* For the moon computations, we need to compute
     * from JDEmoons rather than JDE.
     */
    /* F = NN - (Light Time to Earth in days) */
    //printf("Light time: %lf days (%lf AU)\n", DT/173.83, DT);
    f = JD - JDEmoons - (dt / 173.83);
    f1 = ii + MP - VP;
    //printf("F = %lf, F1 = %lf\n", F, F1);

    /* calc. angle from inferior geocentric conjuction measured westward */
    for (i=0; i < NumMoons; ++i)
    {
        satMoons[i].u = (satMoons[i].u0 + (f * TWOPI / satMoons[i].pd) + f1) / TWOPI;
        satMoons[i].u = (satMoons[i].u - (int)satMoons[i].u) * TWOPI;
    }

    /********************* FIND SATURN'S COORDINATES *********************
     * psi -> Heliocentric Latitude
     * lamb -> Geocentric Ecliptic Longitude
     * beta -> Geocentric Ecliptic Latitude
     */
    psi = asin(sin(lP - Ohm) * sin(IncS));
    lpq = atan2(sin(lP - Ohm) * cos(IncS), cos(lP - Ohm)) + Ohm;
    rpq = rP * cos(psi);
    lamb = atan2(rE * sin(lpq - lE), rpq - rE * cos(lpq - lE)) + lpq;
    beta = rpq * tan(psi) * sin(lamb - lpq);
    beta = atan(beta / (rE * sin(lpq - lE)));
    RA = atan2((sin(lamb) * cos(Obl)) - (tan(beta) * tan(Obl)),
               cos(lamb));
    dec = asin(sin(beta) * cos(Obl) + cos(beta) * sin(Obl) * sin(lamb));
    //printf("Saturn's coordinates are (%f h, %f deg)\n",
    //       RA * 12. / PI, DEC * 180 / PI);

    /***************** FIND INCLINATION OF RINGS *************************
     * Use dot product of Earth-Saturn vector and Saturn's rotation axis
     */
    tva = (90 - 83.51) * D2R; /* Theta coordinate of Saturn's axis */
    pva = 40.27 * D2R;        /* Phi coordinate of Saturn's axis */
    tvc = PI2 - dec;
    pvc = RA;
    dot1 = sin(tva) * cos(pva) * sin(tvc) * cos(pvc)
           + sin(tva) * sin(pva) * sin(tvc) * sin(pvc)
           + cos(tva) * cos(tvc);
    Inclination = acos(dot1);
    if (Inclination > 0) Inclination = PI2 - Inclination;
    else Inclination = -PI2 - Inclination;

    /************** FIND INCLINATION OF IAPETUS' ORBIT *******************
     * Use dot product of Earth-Saturn vector and Iapetus' orbit axis
     * Vector B
     */
    tvb = (90. - 75.6) * D2R;    /* Theta coord of Iapetus' orbit axis (est) */
    pvb = 21.34 * TWOPI / 24;    /* Phi coord of Iapetus' orbit axis (est) */
    dot2 = sin(tvb) * cos(pvb) * sin(tvc) * cos(pvc);
    dot2 = dot2 + sin(tvb) * sin(pvb) * sin(tvc) * sin(pvc);
    dot2 = dot2 + cos(tvb) * cos(tvc);
    IapInci = atan(sqrt(1. - (dot2 * dot2)) / dot2);
       /* satsat said this was arccos, but it's not */
    if (IapInci > 0) IapInci = PI2 - IapInci;
    else IapInci = -PI2 - IapInci;

    /************** FIND ROTATION ANGLE OF IAPETUS' ORBIT ****************
     * Use inclination of Iapetus' orbit with respect to ring plane
     * Triple Product
     */
    trip = sin(tvc) * cos(pvc) * sin(tva) * sin(pva) * cos(tvb);
    trip = trip - sin(tvc) * cos(pvc) * sin(tvb) * sin(pvb) * cos(tva);
    trip = trip + sin(tvc) * sin(pvc) * sin(tvb) * cos(pvb) * cos(tva);
    trip = trip - sin(tvc) * sin(pvc) * sin(tva) * cos(pva) * cos(tvb);
    trip = trip + cos(tvc) * sin(tva) * cos(pva) * sin(tvb) * sin(pvb);
    trip = trip - cos(tvc) * sin(tvb) * cos(pvb) * sin(tva) * sin(pva);
    IapGam = -1 * atan(trip / sqrt(1 - trip * trip));  /* ArcSIN */
       /* satsat said this was arcsin, but it's not */
}

void DrawSaturn(int Xmax, int Ymax, double Scale, int flipNS, int flipEW,
                RGBColorType* skyC, RGBColorType* planetC)
{
    int i;
    double ringAspect = sin(Inclination);    /* Aspect Ratio */
    //double A1, A2;
    double tempX, tempY;
    int X, Y;
    int XC = Xmax / 2;
    int YC = Ymax / 2;
    char Buf[20];

    tempY = Inclination / D2R;
    if (Inclination != 0)
    {
        StrPrintF(Buf, "Inclination: %d.%d",
                  (int)tempY, (int)(Inclination*10./D2R - tempY*10.));
        Message(Buf);
    }

    // Label the cardinal directions:
    ChangeColor(planetC);
    DrawString(XC,      10,     flipNS ? "-N-" : "-S-", 3);
    DrawString(XC,      Ymax-4, flipNS ? "-S-" : "-N-", 3);
    DrawString(Xmax-15, YC,     flipEW ? "-W"  : "-E",  2);
    DrawString(3,       YC,     flipEW ?  "E-" :  "W-", 2);

    SaturnOrbit(julianDate);

    ChangeColor(planetC);

    if (RS * Scale < 3)
    {
        /* Draw Saturn */
        i = 3;
        Line(XC-i,   YC,   XC, YC+i);
        Line(  XC, YC+i, XC+i,   YC);
        Line(XC+i,   YC,   XC, YC-i);
        Line(  XC, YC-i, XC-i,   YC);
        /* and the rings */
        Line(XC-i*2, YC, XC+i*2, YC);
    }
    else
    {
        /* Draw rings */
        Oval(XC, YC, RS4 * Scale + 1, ringAspect, 1);
        ChangeColor(skyC);
        Oval(XC, YC, RS3 * Scale - 1, ringAspect, 1);
        ChangeColor(planetC);
        Oval(XC, YC, RS2 * Scale + 1, ringAspect, 1);
        ChangeColor(skyC);
        Oval(XC, YC, RS1 * Scale - 1, ringAspect, 1);
        ChangeColor(planetC);

        /* Draw planet */
        Oval(XC, YC, RS * Scale, .9, 1);
    }

    /******************* DRAW MOONS (Earth View) *************************/
    for (i=0; i < NumMoons; ++i)
    {
        /* Orbital Paths */
        if (satMoons[i].sma * Scale * ringAspect < Ymax / 4 && DrawOrbits)
            Oval(XC, YC, satMoons[i].sma * Scale, ringAspect, 0);
        X = XC - (satMoons[i].sma * sin(satMoons[i].u) * Scale);
        Y = YC + (satMoons[i].sma * cos(satMoons[i].u) * sin(Inclination) * Scale);
        DrawMoon(flipEW ? Xmax-X : X, flipNS ? Ymax-Y : Y, i);
    }

    /********************* Iapetus' Orbit (Earth View) *******************/
    tempX = -1 * (satMoons[8].sma * sin(satMoons[8].u) * Scale);
    tempY = (int)((satMoons[8].sma * cos(satMoons[8].u) * sin(IapInci) * Scale));
    X = XC + tempX * cos(IapGam) + tempY * sin(IapGam); /* Rotation */
    Y = YC - tempX * sin(IapGam) + tempY * cos(IapGam);
    if (X < Xmax && X >= 0)
    {
        if (Y <= (Ymax / 2 - 10) && Y >= 0)
        {
#ifdef NOTUSED
            /* This code never modifies X and Y! */
            if (O1 != 0)
            {
                /* Orbital Path */
                int tempY, Xlast, Ylast, Xdraw, Ydraw;
                tempY = (int)((satMoons[8].sma * sin(INCI) * Scale));
                Xlast = XC + tempY * sin(IapGam); /* REM Rotation */
                Ylast = YC + tempY * cos(IapGam);
                for (i=0; i<360; i += 2)
                {
                    tempX = -1 * (satMoons[8].sma * sin(I * P) * Scale);
                    tempY = (int)((satMoons[8].sma * cos(I * P) * sin(IapInci) * Scale));
                    Xdraw = XC + tempX * cos(IapGam) + tempY * sin(IapGam);
                        /* REM Rotation */
                    Ydraw = YC - tempX * sin(IapGam) + tempY * cos(IamGam);
                    if (Ydraw < (Ymax / 2 - 10) && Ylast < (Ymax / 2 - 10))
                        Line (Xlast, Ylast, Xdraw, Ydraw);
                    Xlast = Xdraw;
                    Ylast = Ydraw;
                }
             }
#endif /* NOTUSED */
            /* Draw Iapetus */
            DrawMoon(flipEW ? Xmax-X : X, flipNS ? Ymax-Y : Y, 8);
         }
    }
}


