/*
 * xgraphics.c
 * Routines to draw planets, moons and other figures.
 *
 * Copyright 2002 by Akkana Peck.
 * You are free to use or modify this code under the Gnu Public License.
 */

#include "shallow.h"
#include "xgrfx.h"

#include <stdlib.h>    // for getenv
#include <X11/keysym.h>
#include <X11/Xutil.h>

// These are the routines each program has to define:
extern void RedrawScreen();
extern void InitColors();

int XWinSize = 700;
int YWinSize = 300;
double Scale = 0.;
int flipNS = 0;
int flipEW = 0;

Display *display;
int screen;
Window win;
GC gc = 0;

RGBColorType white, black;

double UnixTimeToJulian(time_t sec);

void InitWindow()
{
    if ((display = XOpenDisplay(getenv("DISPLAY"))) == 0)
    {
        fprintf(stderr, "Can't open display: %s\n", getenv("DISPLAY"));
        exit(1);
    }
    screen = DefaultScreen(display);

    win = XCreateSimpleWindow(display, RootWindow(display, screen),
                              0, 0, XWinSize, YWinSize, 3,
                              WhitePixel(display, screen),
                              BlackPixel(display, screen));
    if (!win)
    {
        fprintf(stderr, "Can't create window\n");
        exit(1);
    }

    XSelectInput(display, win,
                 ExposureMask
                 | ButtonPressMask| ButtonReleaseMask
                 | KeyPressMask
                 | StructureNotifyMask
                );

    white = WhitePixel(display, screen);
    black = BlackPixel(display, screen);
    InitColors();
    ClearScreen();
    XMapWindow(display, win);
}

int HandleEvent()
{
    XEvent event;
    char buffer[20];
    KeySym keysym;
    XComposeStatus compose;
    extern double WIDESCALE, NARROWSCALE;
    time_t sec;
    static int lastX = -1;
    static int lastY = -1;

    XNextEvent(display, &event);
    switch (event.type)
    {
      case Expose:
          //printf("Expose\n");
          // Why do we need two of these, you ask?
          // Because we stupidly pass &black and &white in to the
          // routine that draws the screen, before they actually
          // have values.  Sheesh.
          RedrawScreen();
          RedrawScreen();
          break;
      case ConfigureNotify:
          XWinSize = event.xconfigure.width;
          YWinSize = event.xconfigure.height;
          //printf("ConfigureNotify: now (%d, %d)\n", XWinSize, YWinSize);
          break;
      case ButtonPress:
          lastX = event.xbutton.x;
          lastY = event.xbutton.y;
          break;
      case ButtonRelease:
#define MOUSE_THRESHOLD 3
          if (abs(lastX - event.xbutton.x) > MOUSE_THRESHOLD)
          {
              int yc = YWinSize / 2;
              if ((event.xbutton.x > lastX && event.xbutton.y > yc)
                  || (event.xbutton.x < lastX && event.xbutton.y < yc))
                  julianDate -= .5/24.;
              else
                  julianDate += .5/24.;
              RedrawScreen();
          }
          break;
      case KeyPress:
          XLookupString(&(event.xkey), buffer, sizeof buffer,
                        &keysym, &compose);
          switch (keysym)
          {
            case XK_q:
                return -1;
            case XK_t:
            case XK_h:
                time(&sec);
                julianDate = UnixTimeToJulian(sec);
                RedrawScreen();
                break;
            case XK_space:
                julianDate += .24/24;
                RedrawScreen();
                break;
            case XK_minus:
            case XK_BackSpace:
                julianDate -= .24/24;
                RedrawScreen();
                break;
            case XK_slash:
                if (Scale == NARROWSCALE)
                    Scale = WIDESCALE;
                else
                    Scale = NARROWSCALE;
                RedrawScreen();
                break;
            case XK_n:
            case XK_s:
                flipNS = !flipNS;
                RedrawScreen();
                break;
            case XK_e:
            case XK_w:
                flipEW = !flipNS;
                RedrawScreen();
                break;
            default:
                break;
          }
          break;
      case MapNotify:
      case ReparentNotify:
          break;
      default:
          printf("Unknown event: %d\n", event.type);
    }
    return 0;
}

void ChangeColor(RGBColorType* c)
{
    static RGBColorType lastColor = 0;

    if (!gc)
    {
        gc = XCreateGC(display, win,  0, 0);
        if (!gc) {
            printf("ChangeColor: no GC!\n");
            return;
        }
        lastColor = *c + 1;  // anything that's not *c
    }

    if (1 || *c != lastColor)
    {
        XGCValues values;
        values.foreground = *c;
        XChangeGC(display, gc, GCForeground, &values);
        lastColor = *c;
    }
}

void ClearScreen()
{
    ChangeColor(&black);
    // Black the screen
    XFillRectangle(display, win, gc, 0, 0, XWinSize, YWinSize);
}

void DrawString(int x, int y, char* s, int len)
{
    XDrawString(display, win, gc, x, y, s, len);
}

#define FULLCIRCLE 360*64

void Oval(int xc, int yc, int r, double aspect, int fill)
{
    int ra = r * aspect;
    if (fill)
        XFillArc(display, win, gc, xc-r, yc-ra, 2*r, 2*ra, 0, FULLCIRCLE);
    else
        XDrawArc(display, win, gc, xc-r, yc-ra, 2*r, 2*ra, 0, FULLCIRCLE);
}

void Arc(int xc, int yc, int r, double a1, double a2, double aspect, int fill)
{
    int ra = r * aspect;
    if (fill)
        XFillArc(display, win, gc, xc-r, yc-ra, 2*r, 2*ra,
                 (int)(a1 * FULLCIRCLE / TWOPI),
                 (int)(a2 * FULLCIRCLE / TWOPI));
    else
        XDrawArc(display, win, gc, xc-r, yc-ra, 2*r, 2*ra,
                 (int)(a1 * FULLCIRCLE / TWOPI),
                 (int)(a2 * FULLCIRCLE / TWOPI));
}

void Rectangle(int x, int y, int width, int height, int fill)
{
    if (fill)
        XFillRectangle(display, win, gc, x, y, width, height);
    else
        XDrawRectangle(display, win, gc, x, y, width, height);
}

void Line(int x1, int y1, int x2, int y2)
{
    XDrawLine(display, win, gc, x1, y1, x2, y2);
}

void DrawMoon(int x, int y, int num)
{
    char s[2] = { num + '1', 0 };
    //char* s = Roman[num];
    DrawString(x+2, y-2, s, strlen(s));
    Line(x-1, y-1, x-1, y+1);
    Line(x, y-1,   x, y+1);
    Line(x+1, y-1, x+1, y+1);
}

int parseMonth(char* mon)
{
    if (!strncmp(mon, "Jan", 3))
        return 0;
    else if (!strncmp(mon, "Feb", 3))
        return 1;
    else if (!strncmp(mon, "Mar", 3))
        return 2;
    else if (!strncmp(mon, "Apr", 3))
        return 3;
    else if (!strncmp(mon, "May", 3))
        return 4;
    else if (!strncmp(mon, "Jun", 3))
        return 5;
    else if (!strncmp(mon, "Jul", 3))
        return 6;
    else if (!strncmp(mon, "Aug", 3))
        return 7;
    else if (!strncmp(mon, "Sep", 3))
        return 8;
    else if (!strncmp(mon, "Oct", 3))
        return 9;
    else if (!strncmp(mon, "Nov", 3))
        return 10;
    else if (!strncmp(mon, "Dec", 3))
        return 11;

    printf("Can't parse month '%s'\n", mon);
    return 0;
}

